﻿using System;
using System.Collections.Generic;
using System.Drawing;

namespace WindowsFormsApp1
{
    //Třída Environment provádí vykreslování 3D pohledu a logiku prostředí
    class Environment
    {
        //Proměnné prostředí, grafické a kolizní mapy, barvy označující operace, textury
        Bitmap[] map;
        Bitmap[] collisionMap;
        Bitmap frame;
        public Bitmap Map { get { return map[currentMap]; } }
        public Bitmap ColisionMap { get { return collisionMap[currentMap]; } }
        Color free, goal, next, previous;
        Dictionary<Color, Bitmap> textures = new Dictionary<Color, Bitmap>();
        int currentMap = 0;
        double widthCoeficient;
        double refHeight;
        public Point Spawn { get; set; }
        public bool Victory { get; private set; }
        public int ScanDistance { get; private set; }
        int fov = 70;

        //Konstruktor přijímá velikost herního panelu, pole grafických a kolizních map, pole funkčních barev, slovník textur, pozici bodu Spawn, a vzdálenost scanování
        public Environment(Size size, Bitmap[] map, Bitmap[] collisionMap, Color[] functions, Dictionary<Color, Bitmap> textures, Point spawn, int scanDistance = 50)
        {
            //Vykreslovanému snímku je přiřazena velikost herního panelu, a vypočítají se pomocné proměnné pro vykreslování
            frame = new Bitmap(size.Width, size.Height);
            widthCoeficient = frame.Width / 500;
            refHeight = frame.Height * 1.2f;
            this.map = map;
            this.collisionMap = collisionMap;
            //Pole funkčních barev se rozdělí na jednotlivé položky
            free = functions[0];
            goal = functions[1];
            next = functions[2];
            previous = functions[3];
            this.textures = textures;
            Spawn = spawn;
            ScanDistance = scanDistance;
        }

        //Metoda Reset je zavolána při ukončení hry, nastaví všechny proměnné ovlivňující stav hry na výchozí hodnoty aby bylo možno začít novou
        //hru bez nutnosti prostředí znovu inicializovat
        public void Reset()
        {
            currentMap = 0;
            Victory = false;
        }

        //Metoda ChangeMap přepne aktivní mapu pokud je to možné
        private void ChangeMap(int i)
        {
            if (currentMap + i >= 0 && currentMap + i < map.Length) currentMap += i;
        }

        //Metoda Collision je volána hráčem při každém ticku GameTimeru. Pracuje s hodnotami hráče skrze referenční proměnné
        public void Collision(double maxSpeed, ref double xSpeed, ref double ySpeed, ref PointF position)
        {
            //Kontrola je provedena tolikrát kolik je maximální rychlost hráče
            for (int i = 0; i < maxSpeed; i++)
            {
                //K pozici hráče je přičten vektor jeho pohybu vydělený maximální rychlostí, pokud se tedy hráč pohybuje maximální rychlostí,
                //tento vektor dosáhne jenom jeden pixel daleko. Na výsledné pozici je zkontrolována barva pixelu na aktivní kolizní mapě
                Color c = ColisionMap.GetPixel((int)(position.X + xSpeed / maxSpeed), (int)(position.Y + ySpeed / maxSpeed));
                //Pokud nalezená barva odpovídá barvě "goal" hráč je těsně před nárazem do cíle, je tedy vyhlášeno vítězství
                if (c == goal) Victory = true;
                else if (c.G == 71) ChangeMap(1);
                else if (c.B == 71) ChangeMap(-1);
                //Barva "free" se na kolizní mapě označuje stěnu, pokud tedy není detekována, hráč je posunut na kontrolovanou pozici
                if (c != free)
                {
                    position.X += (float)(xSpeed / maxSpeed);
                    position.Y += (float)(ySpeed / maxSpeed);
                }
                //V opačném případě, pokud je na příštím kroku detekována stěna, hráč se zastaví
                else
                {
                    xSpeed = 0;
                    ySpeed = 0;
                    break;
                }
            }
        }

        //Metoda ThreeDee vykresluje trojrozměrnou projekci pohledu hráče z první osoby. Přijímá pozici hráče a úhel jeho natočení
        public Bitmap ThreeDee(Point origin, int angle)
        {
            //Pro kreslení na snímek je používán grafický objekt vytvořený ze snímku
            using (Graphics g = Graphics.FromImage(frame))
            {
                //První krok je vymazat předchozí snímek. Odstranění tohoto řádku vytváří zajímavý efekt, ale zhoršuje hratelnost.
                g.Clear(Color.Transparent);
                //Vlastnosti grafického objektu jsou nastaveny tak, aby nedocházelo k vyhlazování grafiky. Bez tohoto kroku vznikají na snímku čáry
                g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                g.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Half;
                //Snímek je vykreslován po sloupcích. Šířka sloupce je sedmdesátina z pěti set pixelů
                double width = frame.Width / (fov * widthCoeficient);
                //Cyklus se provede sedmdesátkrát, krát kolikrát se do šířky herního pole vejde 500. Při vyšších velikostech herního okna se tedy
                //zvyšuje rozlišení, a také nároky na výkon
                for (int i = 0; i < fov * widthCoeficient; i++)
                {
                    //Zorné pole hráče je nastaveno na 70°, začíná se tedy od -35° a skrze cyklus se dojde až do 35° oproti úhlu natočení hráče
                    double anglenow = angle - (fov / 2) + (double)i / widthCoeficient;
                    PointF vector = new PointF((float)Math.Cos(anglenow * (Math.PI / 180)), (float)Math.Sin(anglenow * (Math.PI / 180)));
                    //Metoda Raycast zjistí vzdálenost a barvu nejbližšího pixelu na grafické mapě v daném směru
                    Color c;
                    double dist = (int)Raycast(origin, vector, out c);
                    //Výška sloupce se vypočítá jako poměr výšky vykreslovacího okna a vzdálenosti objektu který má reprezentovat.
                    //K výpočtu není použita skutečná vzdálenost, pokud by vzdálenost byla přeponou pravoúhlého trojuhelníku, skutečná použitá hodnota
                    //by byla délka strany přilehlé k úhlu u vrcholu hráče. Bez tohoto kroku dochází k "vypouklosti" zorného pole.
                    //Tento postup poskytuje nejlepší reprezentaci pohybu v trojrozměrném prostoru, ale způsobuje "chlupatost" blízkých objektů,
                    //nejspíš kvůli chybě zaokrouhlování
                    double height = refHeight / (1.0 + dist * Math.Cos((i / widthCoeficient - fov/2) * Math.PI / 180) / 20.0);
                    //Pokud je barva nalezená Raycastem klíčem textury, je sloupec vykreslen touto texturou. Pokud ne, vykreslí se barvou "free"
                    //Barva sloupců je také ovlivněna metodami Fade a FadeBMP, které způsobují ztmavnutí vzálených objektů
                    if (textures.ContainsKey(c)) g.DrawImage(FadeBMP(c, (int)dist), (int)width * i, (int)(frame.Height / 2 - height / 2), (int)Math.Ceiling(width), (int)height);
                    else g.FillRectangle(new SolidBrush(Fade(free, (int)dist)), (int)width * i, (int)(frame.Height / 2 - height), (int)Math.Ceiling(width), (int)height * 2);
                    //Pokud jsou v prostředku zorného pole detekovány určité barvy na vzdálenost bližší než hodnota vlastnosti ScanDistance,
                    //zavolá se metoda ChangeMap. Maximální velikost mapy před zvětšením je 50x50 pixelů, pomocí tohoto přepínání map je možné
                    //"zvětšit" herní pole, nebo způsobit iluzi nemožného prostoru
                    if (anglenow == angle && (c == next || (c.G == 71 && c.R != 71)) && dist <= ScanDistance) ChangeMap(1);
                    else if (anglenow == angle && (c == previous || (c.B == 71 && c.R != 71)) && dist <= ScanDistance) ChangeMap(-1);
                }
            }
            return frame;
        }

        //Metoda Raycast podle výchozí pozice a úhlu nachází vzdálenost a barvu nejbližšího pixelu na grafické mapě
        public double Raycast(Point origin, PointF vector, out Color c)
        {
            //Cursor je proměnná uchovávající zkoumanou pozici
            PointF cursor = new PointF(origin.X, origin.Y);
            //Výchozí hodnota out barvy c je černá, zůstane nezměněná pokud Raycast nenajde žádný vhodný pixel v dohledu
            c = Color.Black;
            //Cyklus se provede 250 krát, maximální dohled je tedy 250 pixelů na mapě
            for (int i = 0; i < 250; i++)
            {
                //Metoda GetPixel způsobí výjimku v případě dostane souřadnice mimo obrázek, toto už je ohlídané kontrolou ohraničení při načítání,
                //ale metoda pro jistotu zastaví kdyby měl kurzor přesto vyjet mimo mapu
                if (cursor.X >= 250 || cursor.X <= 0 || cursor.Y >= 250 || cursor.Y <= 0) break;
                //Pokud je nalezen pixel jiné barvy než "free" je jeho barva uložena do out parametru, a cyklus se přeruší
                if (Map.GetPixel((int)cursor.X, (int)cursor.Y) != free && Map.GetPixel((int)cursor.X, (int)cursor.Y).R != 73)
                {
                    c = Map.GetPixel((int)cursor.X, (int)cursor.Y);
                    break;
                }
                //Kurzor se posune směrem vektoru určeného ve volání
                cursor.X += vector.X;
                cursor.Y += vector.Y;
            }
            //Pomocí pytágorovy věty je vypočítána vzdálenost kterou kurzor urazil než našel vhodný pixel, a ta je vrácena
            return Math.Sqrt(Math.Pow(origin.X - cursor.X, 2) + Math.Pow(origin.Y - cursor.Y, 2));
        }

        //Metoda Fade ztmavuje barvu v závislosti na vzdálenosti
        Color Fade(Color c, int dist)
        {
            //Barva je rozložena na složky, z každé z nich je vypočtena dvěstěpadesátina, a množství těchto dílů závislé na vzdálenosti je odečteno
            double rStep = c.R / 250;
            double gStep = c.G / 250;
            double bStep = c.B / 250;
            return Color.FromArgb((int)(c.R - rStep * dist), (int)(c.G - gStep * dist), (int)(c.B - bStep * dist));
        }

        //Metoda FadeBMP ztmavuje texturu v závislosti na vzdálenosti
        Bitmap FadeBMP(Color c, int dist)
        {
            //Metoda používá stejný proces jako metoda Fade, akorát ho aplikuje na každý pixel textury vybrané ze slovníku podle klíče
            Bitmap bmp = new Bitmap(textures[c]);
            for (int i = 0; i < bmp.Width; i++)
            {
                for (int o = 0; o < bmp.Height; o++)
                {
                    Color cl = bmp.GetPixel(i, o);
                    double rStep = (double)cl.R / 250;
                    double gStep = (double)cl.G / 250;
                    double bStep = (double)cl.B / 250;
                    bmp.SetPixel(i, o, Color.FromArgb((int)(cl.R - rStep * dist), (int)(cl.G - gStep * dist), (int)(cl.B - bStep * dist)));
                }
            }
            return bmp;            
        }
    }
}
