﻿using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;

namespace WindowsFormsApp1
{
    //Třída FileHandler se stará o načtení souborů a jejich kontrolu
    class FileHandler
    {
        //Proměnné filehandleru, obrázek ze souboru, řídící barvy, textury, jednotlivé mapy, chyby, pozice bodu spawn, a šířka hráče
        Bitmap image;
        Color free, spawn, goal, next, previous, nothing;
        public Color[] Functions { get { return new Color[] { free, goal, next, previous }; } }
        public Dictionary<Color, Bitmap> Textures { get; set; } = new Dictionary<Color, Bitmap>();
        int mapCount;
        public Bitmap[] Map { get; private set; }
        public Bitmap[] CollisionMap { get; private set; }
        string error = "";
        public string Error { get { return error; } private set { if (error == "") error = value; } }
        Point spawnTL = new Point(0, 0);
        Point spawnBR = new Point(0, 0);
        public Point Spawn { get { return new Point((spawnTL.X + spawnBR.X) / 2, (spawnTL.Y + spawnBR.Y) / 2); } }
        public int Thick { get; private set; }
        
        //Při konstrukci objekt obdrží šířku hráče, výchozí hodnota je 10 pixelů, poté se pokusí načíst a rozkódovat soubor
        public FileHandler(string mapName, int thicc = 10)
        {
            Thick = thicc;
            if (GetMapFile(mapName))
            {
                GetFunctionalColors();
                GetTextures();
                GetMaps(mapCount);
                if (spawnTL.X == 0 && spawnTL.Y == 0 && spawnBR.X == 0 && spawnBR.Y == 0) Error = "No spawnpoint";
            }
        }

        //Metoda GetMapFile načítá soubor MAP.png který by měl být umístěný ve složce spolu s exe souborem
        private bool GetMapFile(string mapName)
        {
            //Pokud soubor neexistuje, nebo nesplňuje formátové požadavky, není možné provést další operace, proto metoda vrátí false a zapíše
            //chybovou hlášku. Setter chyby je nastaven tak, že si zapamatuje jenom první zapsanou chybu. Chyby jsou seřazeny podle závažnosti
            if (File.Exists(mapName))
            {
                image = new Bitmap(Image.FromFile(mapName));
                if (image.Width < 50 || image.Height < 59) Error = "Incorrect map";
                else
                {
                    //Pokud je soubor v pořádku, vypočítá se kolik obsahuje map. Mapa je čtverec o straně 50 pixelů
                    mapCount = image.Width / 50;
                    return true;
                }
            }
            else Error = "Missing map";
            return false;
        }

        //Metoda GetFunctionalColors načítá funkční barvy které mají být definovány na prvních šesti vertikálních pixelech v souboru
        private void GetFunctionalColors()
        {
            Color[] listCheck = new Color[]
            {
                free = image.GetPixel(0, 0),
                spawn = image.GetPixel(0, 1),
                goal = image.GetPixel(0, 2),
                next = image.GetPixel(0, 3),
                previous = image.GetPixel(0, 4),
                nothing = image.GetPixel(0, 5)
            };
            //Pokud funkční barvy nejsou jedinečné, nelze rozlišit jakou funkci mají zastávat, proto je zapsána chyba
            if (listCheck.Distinct().Count() != listCheck.Length) Error = "Duplicate keys";
        }

        //Metoda GetTextures načítá textury které mají být definovány od druhého sloupce na prvních devíti řádcích v souboru. První řádek je klíč
        private void GetTextures()
        {
            int i = 1;
            //Textury se načítají dokud není nalezen okraj souboru, nebo dokud není nalezena funkční barva "nothing" která označuje konec textur
            while (image.GetPixel(i, 0) != nothing && i < image.Width)
            {
                //Textury se ukládají do slovníku, proto jejich klíče nemohou být duplikátní. Pokud jsou, zapíše se chyba a cyklus se přeruší
                if (!Textures.ContainsKey(image.GetPixel(i, 0)))
                {
                    Textures.Add(image.GetPixel(i, 0), image.Clone(new Rectangle(i, 1, 1, 8), System.Drawing.Imaging.PixelFormat.Format32bppArgb));
                    i++;
                }
                else
                {
                    Error = "Duplicate textures";
                    break;
                }
            }
        }

        //Metoda GetMaps inicializuje pole grafických a kolizních map, a plní je pomocí dalších metod
        private void GetMaps(int count)
        {
            Map = new Bitmap[count];
            CollisionMap = new Bitmap[count];
            for (int i = 0; i < count; i++)
            {
                Map[i] = GetMap(i);
                CollisionMap[i] = GetCollisionMap(i);
            }
        }

        //Metoda GetMap překreslí mapu ze souboru na pětinásobek její velikosti, bez vyhlazování. Zvětšení map nám dává více prostoru k pohybu v nich
        private Bitmap GetMap(int index)
        {
            Bitmap bmp = new Bitmap(250, 250);
            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
                g.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Half;
                g.DrawImage(image.Clone(new Rectangle(index * 50, 9, 50, 50), System.Drawing.Imaging.PixelFormat.Format32bppArgb), 0, 0, 250, 250);
            }
            return bmp;
        }

        //Metoda GetCollisionMap podle odpovídající grafické mapy vytvoří kolizní mapu
        private Bitmap GetCollisionMap(int index)
        {
            Bitmap bmp = new Bitmap(250, 250);
            using (Graphics g = Graphics.FromImage(bmp))
            {
                g.Clear(nothing);
                //Metoda cyklicky projde každý pixel grafické mapy
                for (int x = 0; x < Map[index].Width; x++)
                {
                    for (int y = 0; y < Map[index].Height; y++)
                    {
                        //Každý pixel který nemá barvu "free", "goal" nebo "spawn" se počítá jako zeď. Je oklo něj nakreslena kružnice
                        //barvou "free" která se v kolizní mapě jinak nevyskytuje. Velikost kružnice je určena parametrem Thick.
                        //Kolize hráče je počítána pouze podle jednoho bodu, aby se vytvořila iluze že hráč má hloubku a šířku, jsou místo
                        //něj "tlusté" kolizní oblasti stěn
                        if (Map[index].GetPixel(x, y) != free && Map[index].GetPixel(x, y) != goal && Map[index].GetPixel(x, y) != spawn && Map[index].GetPixel(x, y).R != 71 && Map[index].GetPixel(x, y).R != 73)
                        {
                            g.FillEllipse(new SolidBrush(free), new Rectangle(x - Thick / 2, y - Thick / 2, Thick, Thick));
                        }
                        //Pixely barvy "goal" se překreslí přímo. "Goal" by se neměl pokládat příliš blízko ke stěně, jinak je možné že ho
                        //její kolizní oblast překryje, a znemožní tím výhru
                        else if (Map[index].GetPixel(x, y) == goal)
                        {
                            g.FillRectangle(new SolidBrush(goal), new Rectangle(x, y, 1, 1));
                        }
                        else if (Map[index].GetPixel(x, y).R == 71 || Map[index].GetPixel(x, y).R == 73)
                        {
                            g.FillRectangle(new SolidBrush(Map[index].GetPixel(x, y)), new Rectangle(x, y, 1, 1));
                        }
                        //Pokud je nalezen pixel barvy "spawn" na první grafické mapě, zapíší se souřadnice prvního a posledního nalezeného
                        //pixelu této barvy. Tyto pozice při předávání hráči zprůměrují
                        if (index == 0 && Map[index].GetPixel(x, y) == spawn)
                        {
                            Map[index].SetPixel(x, y, free);
                            if (spawnTL.X == 0 && spawnTL.Y == 0) spawnTL = new Point(x, y);
                            else spawnBR = new Point(x, y);
                        }
                        //Pokud je nalezen pixel na okraji mapy který není zdí, zapíše se chyba. Tímto se předchází pohlédnutí, nebo vystoupení
                        //za hranice reality, čímž by byla způsobena výjimka
                        if ((x == 0 || x == Map[index].Width - 1 || y == 0 || y == Map[index].Height - 1) && (Map[index].GetPixel(x, y) == free || Map[index].GetPixel(x, y) == goal || Map[index].GetPixel(x, y) == spawn || Map[index].GetPixel(x, y).R == 71 || Map[index].GetPixel(x, y).R == 73)) Error = "Incomplete borders";
                    }
                }
            }
            return bmp;
        }
    }
}
